# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

import numpy as np

import phonopy
from phonopy import PhonopyGruneisen
from phonopy.interface.calculator import (
    get_calculator_physical_units,
    get_default_cell_filename,
    get_interface_mode,
)


def fracval(frac):
    """Convert fractional value string to float."""
    if frac.find("/") == -1:
        return float(frac)
    else:
        x = frac.split("/")
        return float(x[0]) / float(x[1])


def get_options():
    """Parse command-line options."""
    import argparse

    from phonopy.interface.calculator import (
        add_arguments_of_calculators,
        calculator_info,
    )

    parser = argparse.ArgumentParser(description="Phonopy gruneisen command-line-tool")
    add_arguments_of_calculators(parser, calculator_info)

    parser.set_defaults(
        band_paths=None,
        band_points=51,
        cell_filename=None,
        color_scheme=None,
        cutoff_frequency=None,
        cutoff_wave_vector=1e-4,
        factor=None,
        is_gamma_center=False,
        is_hdf5=False,
        is_nac=False,
        is_mesh_symmetry=True,
        marker="o",
        markersize=None,
        masses=None,
        plot_graph=False,
        primitive_axes=None,
        reads_force_constants=False,
        sampling_mesh=None,
        save_graph=False,
        delta_strain=None,
        supercell_dimension=None,
        symprec=1e-5,
        title=None,
        tmax=2004,
        tmin=0,
        tstep=2,
        volumes_filename=None,
    )
    parser.add_argument(
        "--band", dest="band_paths", help="Band paths in reduced coordinates"
    )
    parser.add_argument(
        "--band_points",
        dest="band_points",
        type=int,
        help="Number of sampling points in a segment of band path",
    )
    parser.add_argument(
        "-c", "--cell", dest="cell_filename", help="Read unit cell", metavar="FILE"
    )
    parser.add_argument("--color", dest="color_scheme", help="Color scheme")
    parser.add_argument(
        "--cutoff",
        dest="cutoff_frequency",
        type=float,
        help="Plot above this cutoff frequency for mesh sampling mode.",
    )
    parser.add_argument(
        "--dim", dest="supercell_dimension", help="Same behavior as DIM tag"
    )
    parser.add_argument(
        "--factor",
        dest="factor",
        type=float,
        help="Conversion factor to favorite frequency unit",
    )
    parser.add_argument(
        "--hdf5",
        dest="is_hdf5",
        action="store_true",
        help="Use hdf5 to read force constants and store results",
    )
    parser.add_argument(
        "--gc",
        "--gamma_center",
        dest="is_gamma_center",
        action="store_true",
        help="Set mesh as Gamma center",
    )
    parser.add_argument("--marker", dest="marker", help="Marker for plot (matplotlib)")
    parser.add_argument(
        "--markersize",
        dest="markersize",
        type=float,
        help="Markersize for plot in points (matplotlib)",
    )
    parser.add_argument("--mass", dest="masses", help="Same as MASS tag")
    parser.add_argument("--mp", "--mesh", dest="sampling_mesh", help="Sampling mesh")
    parser.add_argument(
        "--nac",
        dest="is_nac",
        action="store_true",
        help="Non-analytical term correction",
    )
    parser.add_argument(
        "--nomeshsym",
        dest="is_mesh_symmetry",
        action="store_false",
        help="Symmetry is not imposed for mesh sampling.",
    )
    parser.add_argument(
        "-o", "--output", dest="output_filename", help="Output filename of PDF plot"
    )
    parser.add_argument(
        "-p", "--plot", dest="plot_graph", action="store_true", help="Plot data"
    )
    parser.add_argument(
        "--pa",
        "--primitive-axis",
        "--primitive-axes",
        dest="primitive_axes",
        help="Same as PRIMITIVE_AXES tags",
    )
    parser.add_argument(
        "--q_cutoff",
        dest="cutoff_wave_vector",
        type=float,
        help="Acoustic modes inside cutoff wave vector is treated.",
    )
    parser.add_argument(
        "--readfc",
        dest="reads_force_constants",
        action="store_true",
        help="Read FORCE_CONSTANTS",
    )
    parser.add_argument(
        "-s",
        "--save",
        dest="save_graph",
        action="store_true",
        help="Save plot data in pdf",
    )
    parser.add_argument(
        "--delta-strain",
        dest="delta_strain",
        type=float,
        help="Delta strain instead of using delta-V/V",
    )
    parser.add_argument("-t", "--title", dest="title", help="Title of plot")
    parser.add_argument(
        "--tmax", dest="tmax", type=float, help="Maximum calculated temperature"
    )
    parser.add_argument(
        "--tmin", dest="tmin", type=float, help="Minimum calculated temperature"
    )
    parser.add_argument(
        "--tolerance", dest="symprec", type=float, help="Symmetry tolerance to search"
    )
    parser.add_argument(
        "--tstep", dest="tstep", type=float, help="Calculated temperature step"
    )
    parser.add_argument(
        "--vf",
        "--volumes_filename",
        dest="volumes_filename",
        help="Filename of volume is contained.",
    )
    parser.add_argument(
        "dirnames",
        nargs="*",
        help=("Directory names of phonons with three different volumes (0, +, -)"),
    )
    args = parser.parse_args()
    return args


def run():
    """Run phonopy-gruneisen."""
    args = get_options()

    if len(args.dirnames) != 3:
        sys.stderr.write(
            "Three directory names (original, plus, minus) have to be spefied.\n"
        )
        sys.exit(1)

    primitive_matrix = None
    if args.primitive_axes:
        if args.primitive_axes == "auto":
            primitive_matrix = "auto"
        else:
            vals = [fracval(x) for x in args.primitive_axes.split()]
            primitive_matrix = np.array(vals).reshape(3, 3)

    if args.supercell_dimension:
        supercell_matrix = [int(x) for x in args.supercell_dimension.split()]
        if len(supercell_matrix) == 9:
            supercell_matrix = np.array(supercell_matrix).reshape(3, 3)
        elif len(supercell_matrix) == 3:
            supercell_matrix = np.diag(supercell_matrix)
        else:
            print("Number of elements of --dim option has to be 3 or 9.")
            sys.exit(1)
        if np.linalg.det(supercell_matrix) < 1:
            print("Determinant of supercell matrix has to be positive.")
            sys.exit(1)
    else:
        print("--dim option has to be specified.")
        sys.exit(1)

    #
    # Phonopy gruneisen interface mode
    #
    interface_mode = get_interface_mode(vars(args))
    units = get_calculator_physical_units(interface_mode)

    if args.factor is not None:
        units["factor"] = args.factor

    if args.cell_filename:
        cell_filename = args.cell_filename
    else:
        cell_filename = get_default_cell_filename(interface_mode)

    phonons = []
    for i in range(3):
        directory = args.dirnames[i]

        unitcell_filename = "%s/%s" % (directory, cell_filename)
        print('Unit cell was read from "%s".' % unitcell_filename)

        if args.is_nac:
            born_filename = "%s/BORN" % directory
            print('NAC parameters were read from "%s".' % born_filename)
        else:
            born_filename = None

        if args.reads_force_constants:
            if args.is_hdf5:
                fc_filename = "%s/force_constants.hdf5" % directory
            else:
                fc_filename = "%s/FORCE_CONSTANTS" % directory
            print('Force constants were read from "%s".' % fc_filename)
            phonon = phonopy.load(
                supercell_matrix=supercell_matrix,
                primitive_matrix=primitive_matrix,
                calculator=interface_mode,
                unitcell_filename=unitcell_filename,
                born_filename=born_filename,
                force_constants_filename=fc_filename,
                factor=units["factor"],
                symprec=args.symprec,
            )
            phonons.append(phonon)
        else:
            force_filename = "%s/FORCE_SETS" % directory
            print('Force sets were read from "%s".' % force_filename)
            phonon = phonopy.load(
                supercell_matrix=supercell_matrix,
                primitive_matrix=primitive_matrix,
                calculator=interface_mode,
                unitcell_filename=unitcell_filename,
                born_filename=born_filename,
                force_sets_filename=force_filename,
                factor=units["factor"],
                symprec=args.symprec,
            )
            phonons.append(phonon)

        print("")

    if args.masses:
        masses = [float(v) for v in args.masses.split()]
        for ph in phonons:
            ph.set_masses(masses)

    gruneisen = PhonopyGruneisen(
        phonons[0],  # equilibrium
        phonons[1],  # plus
        phonons[2],  # minus
        delta_strain=args.delta_strain,
    )

    if args.plot_graph:
        if args.save_graph:
            import matplotlib as mpl

            mpl.use("Agg")

    if args.band_paths:
        from phonopy.phonon.band_structure import get_band_qpoints

        band_paths = []
        for path_str in args.band_paths.split(","):
            paths = np.array([fracval(x) for x in path_str.split()])
            if len(paths) % 3 != 0 or len(paths) < 6:
                print("Band path is incorrectly set.")
                sys.exit(1)
            band_paths.append(paths.reshape(-1, 3))
        bands = get_band_qpoints(band_paths, npoints=args.band_points)
        gruneisen.set_band_structure(bands)
        gruneisen.write_yaml_band_structure()
        if args.plot_graph:
            plt = gruneisen.plot_band_structure(
                epsilon=args.cutoff_wave_vector, color_scheme=args.color_scheme
            )
            if args.title is not None:
                plt.suptitle(args.title)

    elif args.sampling_mesh:
        mesh_numbers = np.array([int(x) for x in args.sampling_mesh.split()])
        gruneisen.set_mesh(
            mesh_numbers,
            is_gamma_center=args.is_gamma_center,
            is_mesh_symmetry=args.is_mesh_symmetry,
        )

        if args.is_hdf5:
            gruneisen.write_hdf5_mesh()
        else:
            gruneisen.write_yaml_mesh()

        if args.plot_graph:
            plt = gruneisen.plot_mesh(
                cutoff_frequency=args.cutoff_frequency,
                color_scheme=args.color_scheme,
                marker=args.marker,
                markersize=args.markersize,
            )
            if args.title is not None:
                plt.suptitle(args.title)

        if args.volumes_filename is not None:
            volumes = []
            with open(args.volumes_filename) as f:
                for line in f:
                    if line.strip()[0] == "#":
                        continue
                    volumes.append(float(line.split()[0]))
            if volumes:
                gruneisen.set_thermal_properties(
                    volumes,
                    t_step=args.tstep,
                    t_max=args.tmax,
                    t_min=args.tmin,
                    cutoff_frequency=args.cutoff_frequency,
                )
                gruneisen.write_yaml_thermal_properties()
    else:
        pass

    if args.plot_graph:
        if args.save_graph:
            if args.output_filename:
                plt.savefig(args.output_filename)
            else:
                plt.savefig("gruneisen.pdf")
        else:
            plt.show()



# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Thermal properties plot (propplot)
#
# Usage:
#   propplot

import sys

import numpy as np

try:
    import yaml
except ImportError:
    print("You need to install python-yaml.")
    sys.exit(1)

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy propplot command-line-tool")
    parser.set_defaults(
        output_filename=None,
        factor=1.0,
        is_heat_capacity=False,
        is_entropy=False,
        is_free_energy=False,
        is_diff=False,
        is_gnuplot=False,
        ymax=None,
        ymin=None,
        tmax=None,
        tmin=None,
    )
    parser.add_argument(
        "--factor",
        dest="factor",
        type=float,
        help="Conversion factor of energy unit to internal electronic energy",
    )
    parser.add_argument(
        "--ymax", dest="ymax", type=float, help="Maximum value of y axis"
    )
    parser.add_argument(
        "--ymin", dest="ymin", type=float, help="Minimum value of y axis"
    )
    parser.add_argument(
        "--tmax", dest="tmax", type=float, help="Maximum value of temperature"
    )
    parser.add_argument(
        "--tmin", dest="tmin", type=float, help="Minimum value of temperature"
    )
    parser.add_argument(
        "--cv",
        "--heat_capacity",
        dest="is_heat_capacity",
        action="store_true",
        help="Plot heat capacity",
    )
    parser.add_argument(
        "-s", "--entropy", dest="is_entropy", action="store_true", help="Plot entropy"
    )
    parser.add_argument(
        "--fe",
        "--free_energy",
        dest="is_free_energy",
        action="store_true",
        help="Plot free energy",
    )
    parser.add_argument(
        "-o", "--output", dest="output_filename", help="Output filename"
    )
    parser.add_argument(
        "-d", "--diff", dest="is_diff", action="store_true", help="Calculate difference"
    )
    parser.add_argument(
        "--gnuplot",
        dest="is_gnuplot",
        action="store_true",
        help="Damp thermal properties in gnuplot data style",
    )
    parser.add_argument(
        "filename",
        nargs="*",
        help=("Filename of phonon thermal properties result (thermal_properties.yaml)"),
    )
    args = parser.parse_args()
    return args


def run():
    """Run phonopy-propplot."""
    args = get_options()

    if args.is_gnuplot:
        print(
            "# temperature[K] Free energy [kJ/mol] Entropy [J/K/mol] "
            "Heat capacity [J/K/mol] Energy [kJ/mol]"
        )
    else:
        import matplotlib.pyplot as plt

        if (
            (not args.is_heat_capacity)
            and (not args.is_entropy)
            and (not args.is_free_energy)
        ):
            print("Set --cv, --entropy or --fe")
            sys.exit(0)

    if len(args.filename) == 0:
        filenames = ["thermal_properties.yaml"]
    else:
        filenames = args.filename

    if args.is_heat_capacity:
        prop_target = "heat_capacity"
    elif args.is_entropy:
        prop_target = "entropy"
    elif args.is_free_energy:
        prop_target = "free_energy"

    thermal_properties_0 = yaml.load(open(filenames[0]).read(), Loader=Loader)[
        "thermal_properties"
    ]
    temperatures = [v["temperature"] for v in thermal_properties_0]

    tmin_index = 0
    tmax_index = len(temperatures)
    if args.tmin is not None:
        for i, t in enumerate(temperatures):
            if t > args.tmin - (temperatures[1] - temperatures[0]) * 0.1:
                tmin_index = i
                break

    if args.tmax is not None:
        for i, t in enumerate(temperatures):
            if t > args.tmax + (temperatures[1] - temperatures[0]) * 0.1:
                tmax_index = i
                break

    if args.is_diff:
        props_0 = [v[prop_target] for v in thermal_properties_0]
        props_0 = np.array(props_0) * args.factor

    for filename in filenames:
        thermal_properties = yaml.load(open(filename).read(), Loader=Loader)[
            "thermal_properties"
        ]

        if args.is_gnuplot:
            props = []
            for name in ("temperature", "free_energy", "entropy", "heat_capacity"):
                props.append([v[name] for v in thermal_properties])

            for t, f, e, h in zip(
                props[0][tmin_index:tmax_index],
                props[1][tmin_index:tmax_index],
                props[2][tmin_index:tmax_index],
                props[3][tmin_index:tmax_index],
            ):
                print(("%14.7f " * 5) % (t, f, e, h, f + e * t / 1000))
            print("")
            print("")
        else:
            temperatures = [v["temperature"] for v in thermal_properties]
            props = [v[prop_target] for v in thermal_properties]
            props = np.array(props) * args.factor
            if args.is_diff:
                props -= props_0
            plt.plot(temperatures[tmin_index:tmax_index], props[tmin_index:tmax_index])

    if not args.is_gnuplot:
        if (args.ymin is not None) and (args.ymax is not None):
            plt.ylim(args.ymin, args.ymax)
        elif (args.ymin is None) and (args.ymax is not None):
            plt.ylim(ymax=args.ymax)
        elif (args.ymin is not None) and (args.ymax is None):
            plt.ylim(ymin=args.ymin)

        if args.output_filename is not None:
            plt.rcParams["pdf.fonttype"] = 42
            plt.rcParams["font.family"] = "serif"
            plt.savefig(args.output_filename)
        else:
            plt.show()



"""Tests of file_IO functions."""

import pathlib

import numpy as np
import pytest

import phonopy
from phonopy.file_IO import (
    parse_BORN,
    read_force_constants_hdf5,
    write_force_constants_to_hdf5,
)

cwd = pathlib.Path(__file__).parent
cwd_called = pathlib.Path.cwd()


def test_parse_BORN():
    """Test of parse_BORN."""
    ph = phonopy.load(cwd / "phonopy_disp_NaCl.yaml")
    nac_params = parse_BORN(ph.primitive, filename=cwd / "BORN_NaCl")
    z = np.eye(3) * 1.086875
    epsilon = np.eye(3) * 2.43533967
    np.testing.assert_allclose(nac_params["born"], [z, -z], atol=1e-5)
    np.testing.assert_allclose(nac_params["dielectric"], epsilon, atol=1e-5)
    assert pytest.approx(14.400) == nac_params["factor"]


def test_write_force_constants_to_hdf5():
    """Test write_force_constants_to_hdf5."""
    pytest.importorskip("h5py")

    write_force_constants_to_hdf5(np.zeros(1), physical_unit="eV/angstrom^2")
    for created_filename in ["force_constants.hdf5"]:
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        fc, physical_unit = read_force_constants_hdf5(
            file_path, return_physical_unit=True
        )
        assert fc[0] == pytest.approx(0)
        assert physical_unit == "eV/angstrom^2"
        file_path.unlink()



"""Calculation of density of states."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np

from phonopy.phonon.mesh import Mesh
from phonopy.phonon.tetrahedron_mesh import TetrahedronMesh
from phonopy.structure.tetrahedron_method import TetrahedronMethod


class NormalDistribution:
    """Class to represent normal distribution."""

    def __init__(self, sigma):
        """Init method."""
        self._sigma = sigma

    def calc(self, x):
        """Return normal distribution."""
        return (
            1.0
            / np.sqrt(2 * np.pi)
            / self._sigma
            * np.exp(-(x**2) / 2.0 / self._sigma**2)
        )


class CauchyDistribution:
    """Class to represent Cauchy distribution."""

    def __init__(self, gamma):
        """Init method."""
        self._gamma = gamma

    def calc(self, x):
        """Return Cauchy distribution."""
        return self._gamma / np.pi / (x**2 + self._gamma**2)


class Dos:
    """Base class to calculate density of states."""

    def __init__(self, mesh_object: Mesh, sigma=None, use_tetrahedron_method=False):
        """Init method."""
        self._mesh_object = mesh_object
        self._frequencies = mesh_object.frequencies
        self._weights = mesh_object.weights
        self._tetrahedron_mesh = None
        if use_tetrahedron_method and sigma is None:
            self._tetrahedron_mesh = TetrahedronMesh(
                mesh_object.dynamical_matrix.primitive,
                self._frequencies,
                mesh_object.mesh_numbers,
                np.array(mesh_object.grid_address, dtype="int64"),
                np.array(mesh_object.grid_mapping_table, dtype="int64"),
                mesh_object.ir_grid_points,
            )
        self._frequency_points = None
        self._sigma = sigma
        self.set_draw_area()
        self.set_smearing_function("Normal")

    @property
    def frequency_points(self):
        """Return frequency points."""
        return self._frequency_points

    def set_smearing_function(self, function_name):
        """Set function form for smearing method.

        Parameters
        ----------
        function_name : str
            'Normal': smearing is done by normal distribution.
            'Cauchy': smearing is done by Cauchy distribution.

        """
        if function_name == "Cauchy":
            self._smearing_function = CauchyDistribution(self._sigma)
        else:
            self._smearing_function = NormalDistribution(self._sigma)

    def set_sigma(self, sigma):
        """Set sigma."""
        self._sigma = sigma

    def set_draw_area(self, freq_min=None, freq_max=None, freq_pitch=None):
        """Set frequency points."""
        f_min = self._frequencies.min()
        f_max = self._frequencies.max()

        if self._sigma is None:
            self._sigma = (f_max - f_min) / 100.0

        if freq_min is None:
            f_min -= self._sigma * 10
        else:
            f_min = freq_min

        if freq_max is None:
            f_max += self._sigma * 10
        else:
            f_max = freq_max

        if freq_pitch is None:
            f_delta = (f_max - f_min) / 200.0
        else:
            f_delta = freq_pitch
        self._frequency_points = np.arange(f_min, f_max + f_delta * 0.1, f_delta)


class TotalDos(Dos):
    """Class to calculate total DOS."""

    def __init__(self, mesh_object: Mesh, sigma=None, use_tetrahedron_method=False):
        """Init method."""
        super().__init__(
            mesh_object,
            sigma=sigma,
            use_tetrahedron_method=use_tetrahedron_method,
        )
        self._dos = None
        self._freq_Debye = None
        self._Debye_fit_coef = None
        self._openmp_thm = True

    def run(self):
        """Calculate total DOS."""
        if self._tetrahedron_mesh is None:
            self._dos = np.array(
                [self._get_density_of_states_at_freq(f) for f in self._frequency_points]
            )
        else:
            if self._openmp_thm:
                self._run_tetrahedron_method_dos()
            else:
                self._dos = np.zeros_like(self._frequency_points)
                thm = self._tetrahedron_mesh
                thm.set(value="I", frequency_points=self._frequency_points)
                for i, iw in enumerate(thm):
                    self._dos += np.sum(iw * self._weights[i], axis=1)

    @property
    def dos(self):
        """Return total DOS."""
        return self._dos

    def get_dos(self):
        """Return frequency points and total DOS.

        Returns
        -------
        tuple
            (frequency_points, total_dos)

        """
        warnings.warn(
            "TotalDos.get_dos() is deprecated. "
            "Use frequency_points and dos attributes instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._frequency_points, self._dos

    def get_Debye_frequency(self):
        """Return a kind of Debye frequency."""
        return self._freq_Debye

    def set_Debye_frequency(self, num_atoms, freq_max_fit=None):
        """Calculate a kind of Debye frequency."""
        try:
            from scipy.optimize import curve_fit
        except ImportError as exc:
            raise ModuleNotFoundError("You need to install python-scipy.") from exc

        def Debye_dos(freq, a):
            return a * freq**2

        freq_min = self._frequency_points.min()
        freq_max = self._frequency_points.max()

        if freq_max_fit is None:
            N_fit = int(len(self._frequency_points) / 4.0)  # Hard coded
        else:
            N_fit = int(
                freq_max_fit / (freq_max - freq_min) * len(self._frequency_points)
            )
        popt, pcov = curve_fit(
            Debye_dos, self._frequency_points[0:N_fit], self._dos[0:N_fit]
        )
        a2 = popt[0]
        self._freq_Debye = (3 * 3 * num_atoms / a2) ** (1.0 / 3)
        self._Debye_fit_coef = a2

    def plot(self, ax, xlabel=None, ylabel=None, draw_grid=True, flip_xy=False):
        """Plot total DOS."""
        if flip_xy:
            _xlabel = "Density of states"
            _ylabel = "Frequency"
        else:
            _xlabel = "Frequency"
            _ylabel = "Density of states"

        if xlabel is not None:
            _xlabel = xlabel
        if ylabel is not None:
            _ylabel = ylabel

        plot_total_dos(
            ax,
            self._frequency_points,
            self._dos,
            freq_Debye=self._freq_Debye,
            Debye_fit_coef=self._Debye_fit_coef,
            xlabel=_xlabel,
            ylabel=_ylabel,
            draw_grid=draw_grid,
            flip_xy=flip_xy,
        )

    def write(self, filename="total_dos.dat"):
        """Write total DOS to total_dos.dat."""
        if self._tetrahedron_mesh is None:
            comment = "Sigma = %f" % self._sigma
        else:
            comment = "Tetrahedron method"

        write_total_dos(
            self._frequency_points, self._dos, comment=comment, filename=filename
        )

    def _run_tetrahedron_method_dos(self):
        mesh_numbers = self._mesh_object.mesh_numbers
        cell = self._mesh_object.dynamical_matrix.primitive
        reciprocal_lattice = np.linalg.inv(cell.cell)
        tm = TetrahedronMethod(reciprocal_lattice, mesh=mesh_numbers)
        self._dos = run_tetrahedron_method_dos(
            mesh_numbers,
            self._frequency_points,
            self._frequencies,
            self._mesh_object.grid_address,
            self._mesh_object.grid_mapping_table,
            tm.tetrahedra,
        )

    def _get_density_of_states_at_freq(self, f):
        return np.sum(
            np.dot(self._weights, self._smearing_function.calc(self._frequencies - f))
        ) / np.sum(self._weights)


class ProjectedDos(Dos):
    """Class to calculate projected DOS.

    Attributes
    ----------
    projected_dos : ndarray
        Projected DOS.
        shape=(pdos, frequency_points), dtype="double"
        The first dimension depends on init argument of `xyz_projection`.
        With it True, the length is the number of atoms times 3, otherwise
        the number of atoms.

    """

    def __init__(
        self,
        mesh_object: Mesh,
        sigma=None,
        use_tetrahedron_method=False,
        direction=None,
        xyz_projection=False,
    ):
        """Init method."""
        super().__init__(
            mesh_object,
            sigma=sigma,
            use_tetrahedron_method=use_tetrahedron_method,
        )
        self._eigenvectors = self._mesh_object.eigenvectors
        self._projected_dos = None

        if xyz_projection:
            self._eigvecs2 = np.abs(self._eigenvectors) ** 2
        else:
            num_atom = self._frequencies.shape[1] // 3
            i_x = np.arange(num_atom, dtype="int") * 3
            i_y = np.arange(num_atom, dtype="int") * 3 + 1
            i_z = np.arange(num_atom, dtype="int") * 3 + 2
            if direction is None:
                self._eigvecs2 = np.abs(self._eigenvectors[:, i_x, :]) ** 2
                self._eigvecs2 += np.abs(self._eigenvectors[:, i_y, :]) ** 2
                self._eigvecs2 += np.abs(self._eigenvectors[:, i_z, :]) ** 2
            else:
                d = np.array(direction, dtype="double")
                d /= np.linalg.norm(direction)
                proj_eigvecs = self._eigenvectors[:, i_x, :] * d[0]
                proj_eigvecs += self._eigenvectors[:, i_y, :] * d[1]
                proj_eigvecs += self._eigenvectors[:, i_z, :] * d[2]
                self._eigvecs2 = np.abs(proj_eigvecs) ** 2

        self._openmp_thm = True

    @property
    def partial_dos(self):
        """Return partial DOS."""
        warnings.warn(
            "PartialDos.partial_dos attribute is deprecated. "
            "Use projected_dos attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._projected_dos

    @property
    def projected_dos(self):
        """Return projected DOS."""
        return self._projected_dos

    def run(self):
        """Calculate projected DOS."""
        if self._tetrahedron_mesh is None:
            self._run_smearing_method()
        else:
            if self._openmp_thm:
                self._run_tetrahedron_method_dos()
            else:
                self._run_tetrahedron_method()

    def get_partial_dos(self):
        """Return partial DOS.

        Returns
        -------
        tuple
            frequency_points: Sampling frequencies
            projected_dos: [atom_index, frequency_points_index]

        """
        warnings.warn(
            "ProjectedDos.get_partial_dos() is deprecated. "
            "Use frequency_points and projected_dos attributes instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._frequency_points, self._projected_dos

    def plot(
        self,
        ax,
        indices=None,
        legend=None,
        legend_prop=None,
        legend_frameon=True,
        xlabel=None,
        ylabel=None,
        draw_grid=True,
        flip_xy=False,
    ):
        """Plot projected DOS."""
        if flip_xy:
            _xlabel = "Partial density of states"
            _ylabel = "Frequency"
        else:
            _xlabel = "Frequency"
            _ylabel = "Partial density of states"

        if xlabel is not None:
            _xlabel = xlabel
        if ylabel is not None:
            _ylabel = ylabel

        plot_projected_dos(
            ax,
            self._frequency_points,
            self._projected_dos,
            indices=indices,
            legend=legend,
            legend_prop=legend_prop,
            legend_frameon=legend_frameon,
            xlabel=_xlabel,
            ylabel=_ylabel,
            draw_grid=draw_grid,
            flip_xy=flip_xy,
        )

    def write(self, filename="projected_dos.dat"):
        """Write projected DOS to projected_dos.dat."""
        if self._tetrahedron_mesh is None:
            comment = "Sigma = %f" % self._sigma
        else:
            comment = "Tetrahedron method"

        write_projected_dos(
            self._frequency_points,
            self._projected_dos,
            comment=comment,
            filename=filename,
        )

    def _run_smearing_method(self):
        num_pdos = self._eigvecs2.shape[1]
        num_freqs = len(self._frequency_points)
        self._projected_dos = np.zeros((num_pdos, num_freqs), dtype="double")
        weights = self._weights / float(np.sum(self._weights))
        for i, freq in enumerate(self._frequency_points):
            amplitudes = self._smearing_function.calc(self._frequencies - freq)
            for j in range(self._projected_dos.shape[0]):
                self._projected_dos[j, i] = np.dot(
                    weights, self._eigvecs2[:, j, :] * amplitudes
                ).sum()

    def _run_tetrahedron_method(self):
        num_pdos = self._eigvecs2.shape[1]
        num_freqs = len(self._frequency_points)
        self._projected_dos = np.zeros((num_pdos, num_freqs), dtype="double")
        thm = self._tetrahedron_mesh
        thm.set(value="I", frequency_points=self._frequency_points)
        for i, iw in enumerate(thm):
            w = self._weights[i]
            self._projected_dos += np.dot(iw * w, self._eigvecs2[i].T).T

    def _run_tetrahedron_method_dos(self):
        mesh_numbers = self._mesh_object.mesh_numbers
        cell = self._mesh_object.dynamical_matrix.primitive
        reciprocal_lattice = np.linalg.inv(cell.cell)
        tm = TetrahedronMethod(reciprocal_lattice, mesh=mesh_numbers)
        pdos = run_tetrahedron_method_dos(
            mesh_numbers,
            self._frequency_points,
            self._frequencies,
            self._mesh_object.grid_address,
            self._mesh_object.grid_mapping_table,
            tm.tetrahedra,
            coef=self._eigvecs2,
        )
        self._projected_dos = pdos.T


class PartialDos(ProjectedDos):
    """Class to calculate partial DOS."""

    def __init__(
        self,
        mesh_object: Mesh,
        sigma=None,
        use_tetrahedron_method=False,
        direction=None,
        xyz_projection=False,
    ):
        """Init method."""
        warnings.warn(
            "PartialDos class is deprecated. Use ProjectedDOS instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(
            mesh_object,
            sigma=sigma,
            use_tetrahedron_method=use_tetrahedron_method,
            direction=direction,
            xyz_projection=xyz_projection,
        )


def get_pdos_indices(symmetry):
    """Return atomic indieces grouped by symmetry."""
    mapping = symmetry.get_map_atoms()
    return [list(np.where(mapping == i)[0]) for i in symmetry.get_independent_atoms()]


def write_total_dos(
    frequency_points, total_dos, comment=None, filename="total_dos.dat"
):
    """Write total_dos.dat."""
    with open(filename, "w") as fp:
        if comment is not None:
            fp.write("# %s\n" % comment)

        for freq, dos in zip(frequency_points, total_dos):
            fp.write("%20.10f%20.10f\n" % (freq, dos))


def write_partial_dos(
    frequency_points, partial_dos, comment=None, filename="partial_dos.dat"
):
    """Write partial_dos.dat."""
    warnings.warn(
        "write_partial_dos() is deprecated. Use write_projected_dos() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    write_projected_dos(
        frequency_points, partial_dos, comment=comment, filename=filename
    )


def write_projected_dos(
    frequency_points, projected_dos, comment=None, filename="projected_dos.dat"
):
    """Write projected_dos.dat."""
    with open(filename, "w") as fp:
        if comment is not None:
            fp.write("# %s\n" % comment)

        for freq, pdos in zip(frequency_points, projected_dos.T):
            fp.write("%20.10f" % freq)
            fp.write(("%20.10f" * len(pdos)) % tuple(pdos))
            fp.write("\n")


def plot_total_dos(
    ax,
    frequency_points,
    total_dos,
    freq_Debye=None,
    Debye_fit_coef=None,
    xlabel=None,
    ylabel=None,
    draw_grid=True,
    flip_xy=False,
):
    """Plot total DOS."""
    ax.xaxis.set_ticks_position("both")
    ax.yaxis.set_ticks_position("both")
    ax.xaxis.set_tick_params(which="both", direction="in")
    ax.yaxis.set_tick_params(which="both", direction="in")

    if freq_Debye is not None:
        freq_pitch = frequency_points[1] - frequency_points[0]
        num_points = int(freq_Debye / freq_pitch)
        freqs = np.linspace(0, freq_Debye, num_points + 1)

    if flip_xy:
        ax.plot(total_dos, frequency_points, "r-", linewidth=1)
        if freq_Debye:
            ax.plot(
                np.append(Debye_fit_coef * freqs**2, 0),
                np.append(freqs, freq_Debye),
                "b-",
                linewidth=1,
            )
    else:
        ax.plot(frequency_points, total_dos, "r-", linewidth=1)
        if freq_Debye:
            ax.plot(
                np.append(freqs, freq_Debye),
                np.append(Debye_fit_coef * freqs**2, 0),
                "b-",
                linewidth=1,
            )

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    ax.grid(draw_grid)


def plot_partial_dos(
    ax,
    frequency_points,
    partial_dos,
    indices=None,
    legend=None,
    legend_prop=None,
    legend_frameon=True,
    xlabel=None,
    ylabel=None,
    draw_grid=True,
    flip_xy=False,
):
    """Plot partial DOS."""
    warnings.warn(
        "plot_partial_dos() is deprecated. Use plot_projected_dos() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    plot_projected_dos(
        ax,
        frequency_points,
        partial_dos,
        indices=indices,
        legend=legend,
        legend_prop=legend_prop,
        legend_frameon=legend_frameon,
        xlabel=xlabel,
        ylabel=ylabel,
        draw_grid=draw_grid,
        flip_xy=flip_xy,
    )


def plot_projected_dos(
    ax,
    frequency_points,
    projected_dos,
    indices=None,
    legend=None,
    legend_prop=None,
    legend_frameon=True,
    xlabel=None,
    ylabel=None,
    draw_grid=True,
    flip_xy=False,
):
    """Plot projected DOS."""
    ax.xaxis.set_ticks_position("both")
    ax.yaxis.set_ticks_position("both")
    ax.xaxis.set_tick_params(which="both", direction="in")
    ax.yaxis.set_tick_params(which="both", direction="in")

    plots = []
    num_pdos = len(projected_dos)

    if indices is None:
        indices = []
        for i in range(num_pdos):
            indices.append([i])

    for set_for_sum in indices:
        pdos_sum = np.zeros_like(frequency_points)
        for i in set_for_sum:
            if i > num_pdos - 1:
                print("Index number '%d' is specified," % (i + 1))
                print("but it is not allowed to be larger than the number of atoms.")
                raise ValueError
            if i < 0:
                print(
                    "Index number '%d' is specified, but it must be positive." % (i + 1)
                )
                raise ValueError
            pdos_sum += projected_dos[i]
        if flip_xy:
            plots.append(ax.plot(pdos_sum, frequency_points, linewidth=1))
        else:
            plots.append(ax.plot(frequency_points, pdos_sum, linewidth=1))

    if legend is not None:
        ax.legend(legend, prop=legend_prop, frameon=legend_frameon)

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    ax.grid(draw_grid)


def run_tetrahedron_method_dos(
    mesh,
    frequency_points,
    frequencies,
    grid_address,
    grid_mapping_table,
    relative_grid_address,
    coef=None,
):
    """Return (P)DOS calculated by tetrahedron method in C."""
    try:
        import phonopy._phonopy as phonoc
    except ImportError as exc:
        raise RuntimeError("Phonopy C-extension has to be built properly.") from exc

    if coef is None:
        _coef = np.ones((frequencies.shape[0], 1, frequencies.shape[1]), dtype="double")
    else:
        _coef = np.array(coef, dtype="double", order="C")
    arr_shape = frequencies.shape + (len(frequency_points), _coef.shape[1])
    dos = np.zeros(arr_shape, dtype="double")

    phonoc.tetrahedron_method_dos(
        dos,
        np.array(mesh, dtype="int64"),
        frequency_points,
        frequencies,
        _coef,
        np.array(grid_address, dtype="int64", order="C"),
        np.array(grid_mapping_table, dtype="int64", order="C"),
        relative_grid_address,
    )
    if coef is None:
        return dos[:, :, :, 0].sum(axis=0).sum(axis=0) / np.prod(mesh)
    else:
        return dos.sum(axis=0).sum(axis=0) / np.prod(mesh)


def get_dos_frequency_range(freqs, dos):
    """Return reasonable frequency range."""
    i_min = 0
    i_max = 1000

    for i, (_, d) in enumerate(zip(freqs, dos)):
        if d > 1e-5:
            i_min = i
            break

    for i, (_, d) in enumerate(zip(freqs[::-1], dos[::-1])):
        if d > 1e-5:
            i_max = len(freqs) - 1 - i
            break

    f_min = freqs[i_min]
    if f_min > 0:
        f_min = 0

    f_max = freqs[i_max]
    f_max += (f_max - f_min) * 0.05

    return f_min, f_max



"""MLP interfaces."""

# Copyright (C) 2024 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import os
from typing import Any, Optional, Union

import numpy as np

from phonopy.interface.pypolymlp import (
    PypolymlpParams,
    develop_mlp_by_pypolymlp,
    evalulate_pypolymlp,
    load_pypolymlp,
    save_pypolymlp,
)
from phonopy.structure.atoms import PhonopyAtoms


class PhonopyMLP:
    """PhonopyMLP class."""

    def __init__(self, mlp: Optional[Any] = None, log_level: int = 0):
        self._mlp = mlp
        self._log_level = log_level

    @property
    def mlp(self) -> Any:
        """Return MLP instance."""
        return self._mlp

    def save(self, filename: Optional[str] = None):
        """Save MLP."""
        if filename is None:
            _filename = "polymlp.yaml"
        else:
            _filename = filename
        save_pypolymlp(self._mlp, _filename)

    def load(
        self, filename: Optional[Union[str, bytes, os.PathLike]] = None
    ) -> PhonopyMLP:
        """Load MLP."""
        if filename is None:
            _filename = "polymlp.yaml"
        else:
            _filename = filename
        self._mlp = load_pypolymlp(_filename)
        return self

    def evaluate(
        self, supercells_with_displacements: list[PhonopyAtoms]
    ) -> list[np.ndarray, np.ndarray, np.ndarray]:
        """Evaluate MLP."""
        return evalulate_pypolymlp(self._mlp, supercells_with_displacements)

    def develop(
        self,
        mlp_dataset: dict,
        supercell: PhonopyAtoms,
        params: Optional[Union[PypolymlpParams, dict, str]] = None,
        test_size: float = 0.1,
    ):
        """Develop MLP."""
        self._mlp = develop_mlp_by_pypolymlp(
            mlp_dataset,
            supercell,
            params=params,
            test_size=test_size,
            log_level=self._log_level,
        )



"""Tests for ABACUS calculater interface."""

import os

import numpy as np

from phonopy.interface.abacus import read_abacus, read_abacus_output
from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.structure import cells

data_dir = os.path.dirname(os.path.abspath(__file__))


def test_read_abacus_nomag():
    """Test of read_ABACUS."""
    cell, pps, orbitals, abfs = read_abacus(os.path.join(data_dir, "NaCl-abacus.stru"))
    filename = os.path.join(data_dir, "NaCl-abinit-pwscf.yaml")
    cell_ref = read_cell_yaml(filename)
    # assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    assert cells.isclose(cell, cell_ref)
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r
    assert pps["Na"] == "Na_ONCV_PBE-1.0.upf"
    assert pps["Cl"] == "Cl_ONCV_PBE-1.0.upf"
    assert orbitals["Na"] == "Na_gga_9au_100Ry_4s2p1d.orb"
    assert orbitals["Cl"] == "Cl_gga_8au_100Ry_2s2p1d.orb"


def test_read_abacus_mag():
    """Test of read_ABACUS with magnetic moments."""
    cell, pps, orbitals, abfs = read_abacus(
        os.path.join(data_dir, "NaCl-abacus-mag.stru")
    )
    filename = os.path.join(data_dir, "NaCl-abacus-mag.yaml")
    cell_ref = read_cell_yaml(filename)
    # assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    assert cells.isclose(cell, cell_ref)
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r
    assert pps["Na"] == "Na_ONCV_PBE-1.0.upf"
    assert pps["Cl"] == "Cl_ONCV_PBE-1.0.upf"
    assert orbitals["Na"] == "Na_gga_9au_100Ry_4s2p1d.orb"
    assert orbitals["Cl"] == "Cl_gga_8au_100Ry_2s2p1d.orb"

    diff_mag = cell_ref.magnetic_moments - np.array([1] * 4 + [2] * 4)
    assert (np.abs(diff_mag) < 1e-5).all()


def test_read_abacus_mag_noncolin():
    """Test of read_ABACUS with magnetic moments."""
    cell, pps, orbitals, abfs = read_abacus(
        os.path.join(data_dir, "NaCl-abacus-mag-noncolin.stru")
    )
    filename = os.path.join(data_dir, "NaCl-abacus-mag-noncolin.yaml")
    cell_ref = read_cell_yaml(filename)
    # assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    assert cells.isclose(cell, cell_ref)
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r
    assert pps["Na"] == "Na_ONCV_PBE-1.0.upf"
    assert pps["Cl"] == "Cl_ONCV_PBE-1.0.upf"
    assert orbitals["Na"] == "Na_gga_9au_100Ry_4s2p1d.orb"
    assert orbitals["Cl"] == "Cl_gga_8au_100Ry_2s2p1d.orb"
    diff_mag = cell_ref.magnetic_moments - cell.magnetic_moments
    assert (np.abs(diff_mag) < 1e-5).all()


def test_read_abacus_output():
    """Test of read abacus output."""
    force = read_abacus_output(os.path.join(data_dir, "NaCl-abacus.out"))
    assert force.mean() < 1e-10
    assert force[0][0] + 1.85537138e-02 < 1e-5



"""Group velocity example by NaCl."""

import matplotlib.pyplot as plt
import numpy as np

import phonopy

phonon = phonopy.load(
    unitcell_filename="POSCAR-unitcell",
    born_filename="BORN",
    force_sets_filename="FORCE_SETS",
    supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
)
phonon.run_mesh([31, 31, 31], with_group_velocities=True)
phonon_mesh = phonon.get_mesh_dict()
frequencies = phonon_mesh["frequencies"]
group_velocity = phonon_mesh["group_velocities"]
gv_norm = np.sqrt((group_velocity**2).sum(axis=2))
for i, (f, g) in enumerate(zip(frequencies.T, gv_norm.T)):
    plt.plot(f, g, "o", label=("band%d" % (i + 1)))
plt.legend()
plt.xlabel("Frequency (THz)")
plt.ylabel("|group-velocity| (THz.A)")
plt.show()



"""phonopy.yaml reader and writer."""

# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import dataclasses
import io
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

import numpy as np
import yaml

from phonopy.structure.cells import Primitive, Supercell
from phonopy.structure.symmetry import Symmetry

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

if TYPE_CHECKING:
    from phonopy import Phonopy

from phonopy.file_IO import get_io_module_to_decompress
from phonopy.structure.atoms import PhonopyAtoms, parse_cell_dict
from phonopy.structure.dataset import forces_in_dataset


@dataclasses.dataclass
class PhonopyYamlData:
    """PhonopyYaml data structure."""

    configuration: Optional[dict] = None
    calculator: Optional[str] = None
    physical_units: Optional[dict] = None
    unitcell: Optional[PhonopyAtoms] = None
    primitive: Optional[Union[Primitive, PhonopyAtoms]] = None
    supercell: Optional[Union[Supercell, PhonopyAtoms]] = None
    dataset: Optional[dict] = None
    supercell_matrix: Optional[np.ndarray] = None
    primitive_matrix: Optional[np.ndarray] = None
    nac_params: Optional[dict] = None
    force_constants: Optional[np.ndarray] = None
    symmetry: Optional[Symmetry] = None  # symmetry of supercell
    frequency_unit_conversion_factor: Optional[float] = None
    version: Optional[str] = None
    command_name: str = "phonopy"


def phonopy_yaml_property_factory(name):
    """Property factor for PhonopyYaml class."""

    def getter(instance):
        return instance._data.__dict__[name]

    def setter(instance, value):
        instance._data.__dict__[name] = value

    return property(getter, setter)


class PhonopyYamlLoaderBase(ABC):
    """Base class of PhonopyYaml loader."""

    def __init__(
        self, yaml_data, configuration=None, calculator=None, physical_units=None
    ):
        """Init method.

        Parameters
        ----------
        yaml_data : dict
            Data in dict, which is normally obtained parsing ``phonopy.yaml``
            like file using pyyaml.
        configuration : dict, optional
            Phonopy calculation configuration. Default is None.
        calculator : str, optional
            Force calculator. Default is None.
        physical_units : dict, optional
            Physical units used for the phonopy calculation.

        """
        self._yaml = yaml_data
        self._data = PhonopyYamlData(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
        )

    @property
    def data(self):
        """Return PhonopyYamlData instance."""
        return self._data

    def parse(self):
        """Parse raw yaml data."""
        self._parse_command_header()
        self._parse_physical_units()
        self._parse_transformation_matrices()
        self._parse_all_cells()
        self._parse_force_constants()
        self._parse_dataset()
        self._parse_nac()
        self._parse_calculator()
        return self

    def _parse_command_header(self):
        if self._data.command_name in self._yaml:
            if self._data.command_name in self._yaml:
                header = self._yaml[self._data.command_name]
                if "version" in header:
                    self._data.version = header["version"]

    def _parse_physical_units(self):
        if "physical_unit" in self._yaml:
            self._data.physical_units = {}
            for key, val in self._yaml["physical_unit"].items():
                if key == "atomic_mass":
                    continue
                if key in ["length", "force", "force_constants"]:
                    self._data.physical_units[key + "_unit"] = val
                else:
                    self._data.physical_units[key] = val

    def _parse_transformation_matrices(self):
        if "supercell_matrix" in self._yaml:
            self._data.supercell_matrix = np.array(
                self._yaml["supercell_matrix"], dtype="intc", order="C"
            )
        if "primitive_matrix" in self._yaml:
            self._data.primitive_matrix = np.array(
                self._yaml["primitive_matrix"], dtype="double", order="C"
            )

    def _parse_all_cells(self):
        if "unit_cell" in self._yaml:
            self._data.unitcell = self._parse_cell(self._yaml["unit_cell"])
        if "primitive_cell" in self._yaml:
            self._data.primitive = self._parse_cell(self._yaml["primitive_cell"])
        if "supercell" in self._yaml:
            self._data.supercell = self._parse_cell(self._yaml["supercell"])
        if self._data.unitcell is None:
            if "lattice" in self._yaml and (
                "points" in self._yaml or "atoms" in self._yaml
            ):
                self._data.unitcell = parse_cell_dict(self._yaml)

    def _parse_cell(self, cell_dict):
        return parse_cell_dict(cell_dict)

    def _parse_force_constants(self):
        if "force_constants" in self._yaml:
            shape = tuple(self._yaml["force_constants"]["shape"]) + (3, 3)
            fc = np.reshape(self._yaml["force_constants"]["elements"], shape)
            self._data.force_constants = np.array(fc, dtype="double", order="C")

    @abstractmethod
    def _parse_dataset(self):
        pass

    def _get_dataset(
        self, supercell: PhonopyAtoms, key_prefix: str = ""
    ) -> Optional[dict]:
        dataset = None
        if f"{key_prefix}displacements" in self._yaml:
            if supercell is not None:
                natom = len(supercell)
            else:
                natom = None

            disp = self._yaml[f"{key_prefix}displacements"][0]
            if isinstance(disp, dict):  # type1
                dataset = self._parse_force_sets_type1(
                    natom=natom, key_prefix=key_prefix
                )
            elif isinstance(disp, list):  # type2
                if "displacement" in disp[0]:
                    dataset = self._parse_force_sets_type2_v223(key_prefix=key_prefix)
        elif f"{key_prefix}dataset" in self._yaml:
            dataset = self._parse_force_sets_type2(key_prefix=key_prefix)
        return dataset

    def _parse_force_sets_type1(
        self, natom: Optional[int] = None, key_prefix: str = ""
    ) -> dict:
        key = f"{key_prefix}displacements"
        if "forces" in self._yaml[key][0]:
            dataset = {"natom": len(self._yaml[key][0]["forces"])}
        elif natom is not None:
            dataset = {"natom": natom}
        elif "natom" in self._yaml:
            dataset = {"natom": self._yaml["natom"]}
        else:
            raise RuntimeError("Number of atoms in supercell could not be found.")

        first_atoms = []
        for d in self._yaml[key]:
            data = {
                "number": d["atom"] - 1,
                "displacement": np.array(d["displacement"], dtype="double"),
            }
            if "forces" in d:
                data["forces"] = np.array(d["forces"], dtype="double", order="C")
            if "supercell_energy" in d:
                data["supercell_energy"] = d["supercell_energy"]
            first_atoms.append(data)
        dataset["first_atoms"] = first_atoms

        return dataset

    def _parse_force_sets_type2(self, key_prefix: str = "") -> dict:
        """Parse displacements, forces and energies in type2 format.

        This is the format >= v2.24.0 as follows:

        dataset:
          displacements:
            ...
          forces:
            ...
          supercell_energies:
            ...

        """
        key = f"{key_prefix}dataset"
        dataset = {}
        if "displacements" in self._yaml[key]:
            dataset["displacements"] = np.array(
                self._yaml[key]["displacements"], dtype="double", order="C"
            )
        if "forces" in self._yaml[key]:
            dataset["forces"] = np.array(
                self._yaml[key]["forces"], dtype="double", order="C"
            )
        if "supercell_energies" in self._yaml[key]:
            dataset["supercell_energies"] = np.array(
                self._yaml[key]["supercell_energies"], dtype="double"
            )
        return dataset

    def _parse_force_sets_type2_v223(self, key_prefix: str = "") -> dict:
        """Parse displacements, forces and energies in type2 legacy format.

        This is the format < v2.24 as follows:

        displacements:
        - # 1
          - displacement: # 1
              [  -0.0201336051051884,  0.0137526506253330,  0.0174786311319239 ]
              force:
              [   0.0551556600000000, -0.0257346500000000, -0.0282983400000000 ]
          ...
        ...
        supercell_energies:
        - -216.84472784 # 1
        ...

        """
        key = f"{key_prefix}displacements"
        nsets = len(self._yaml[key])
        natom = len(self._yaml[key][0])
        if "force" in self._yaml[key][0][0]:
            with_forces = True
            forces = np.zeros((nsets, natom, 3), dtype="double", order="C")
        else:
            with_forces = False
        displacements = np.zeros((nsets, natom, 3), dtype="double", order="C")
        for i, dfset in enumerate(self._yaml[key]):
            for j, df in enumerate(dfset):
                if with_forces:
                    forces[i, j] = df["force"]
                displacements[i, j] = df["displacement"]

        if with_forces:
            dataset = {"forces": forces, "displacements": displacements}
        else:
            dataset = {"displacements": displacements}

        if "supercell_energies" in self._yaml:
            dataset["supercell_energies"] = np.array(
                self._yaml[f"{key_prefix}supercell_energies"], dtype="double"
            )

        return dataset

    def _parse_nac(self):
        """Parse NAC parameters.

        Older than v2.18, keys below "nac" are put on top level.

        nac:
          born_effective_charge:
            ...
          dielectric_constant:
            ...
          unit_conversion_factor:
            ...
          method:
            ...

        """
        nac_params = self._parse_nac_params(self._yaml)  # older than v2.18
        if "nac" in self._yaml:
            nac_params = self._parse_nac_params(self._yaml["nac"])
        if "born" in nac_params and "dielectric" in nac_params:
            self._data.nac_params = nac_params

    def _parse_nac_params(self, nac_yaml):
        nac_params = {}
        if "born_effective_charge" in nac_yaml:
            nac_params["born"] = np.array(
                nac_yaml["born_effective_charge"], dtype="double", order="C"
            )
        if "dielectric_constant" in nac_yaml:
            nac_params["dielectric"] = np.array(
                nac_yaml["dielectric_constant"], dtype="double", order="C"
            )
        if (  # older than v2.18
            self._data.command_name in nac_yaml
            and "nac_unit_conversion_factor" in nac_yaml[self._data.command_name]
        ):
            nac_params["factor"] = nac_yaml[self._data.command_name][
                "nac_unit_conversion_factor"
            ]
        if "unit_conversion_factor" in nac_yaml:
            nac_params["factor"] = nac_yaml["unit_conversion_factor"]
        if "nac" in self._yaml and "method" in nac_yaml:
            nac_params["method"] = nac_yaml["method"].lower()
        return nac_params

    def _parse_calculator(self):
        if (
            self._data.command_name in self._yaml
            and "calculator" in self._yaml[self._data.command_name]
        ):
            self._data.calculator = self._yaml[self._data.command_name]["calculator"]


class PhonopyYamlLoader(PhonopyYamlLoaderBase):
    """PhonopyYaml loader."""

    def _parse_dataset(self):
        self._data.dataset = self._get_dataset(self._data.supercell)


class PhonopyYamlDumperBase(ABC):
    """Base class of PhonopyYaml dumper."""

    _default_dumper_settings = {
        "force_sets": True,
        "displacements": True,
        "force_constants": False,
        "born_effective_charge": True,
        "dielectric_constant": True,
    }

    def __init__(self, data: PhonopyYamlData, dumper_settings=None):
        """Init method."""
        self._data = data
        self._init_dumper_settings(dumper_settings)

    def get_yaml_lines(self):
        """Return yaml string lines as a list."""
        lines = self._header_yaml_lines()
        lines += self._physical_units_yaml_lines()
        lines += self._symmetry_yaml_lines()
        lines += self._cell_info_yaml_lines()
        lines += self._nac_yaml_lines()
        lines += self._dataset_yaml_lines()
        lines += self._force_constants_yaml_lines()
        return lines

    def _header_yaml_lines(self):
        lines = []
        lines.append("%s:" % self._data.command_name)
        if self._data.version is None:
            from phonopy.version import __version__

            version = __version__
        else:
            version = self._data.version
        lines.append(f'  version: "{version}"')
        if self._data.calculator:
            lines.append("  calculator: %s" % self._data.calculator)
        if self._data.frequency_unit_conversion_factor:
            lines.append(
                "  frequency_unit_conversion_factor: %f"
                % self._data.frequency_unit_conversion_factor
            )
        if self._data.symmetry:
            lines.append("  symmetry_tolerance: %.5e" % self._data.symmetry.tolerance)
        if self._data.configuration is not None:
            lines.append("  configuration:")
            for key in self._data.configuration:
                val = self._data.configuration[key]
                if isinstance(val, str):
                    val = val.replace("\\", "\\\\")
                lines.append('    %s: "%s"' % (key, val))
        lines.append("")
        return lines

    def _physical_units_yaml_lines(self):
        lines = []
        units = self._data.physical_units
        if units is not None:
            lines.append("physical_unit:")
            lines.append('  atomic_mass: "AMU"')
            length_unit = units.get("length_unit", None)
            if length_unit is not None:
                lines.append(f'  length: "{length_unit}"')
            force_unit = units.get("force_unit", None)
            if force_unit is not None and forces_in_dataset(self._data.dataset):
                lines.append(f'  force: "{force_unit}"')
            if self._data.command_name == "phonopy":
                fc_unit = units.get("force_constants_unit", None)
                if fc_unit is not None and self._data.force_constants is not None:
                    lines.append(f'  force_constants: "{fc_unit}"')
            lines.append("")
        return lines

    def _symmetry_yaml_lines(self):
        lines = []
        if self._data.symmetry is None:
            return lines

        dataset = self._data.symmetry.dataset
        if dataset is not None:
            try:
                uni_number = dataset.uni_number
                lines.append("magnetic_space_group:")
                lines.append(f"  uni_number: {uni_number}")
                lines.append(f"  msg_type: {dataset.msg_type}")
                lines.append("")
            except AttributeError:
                lines.append("space_group:")
                spg_type = dataset.international
                lines.append(f'  type: "{spg_type}"')
                lines.append(f"  number: {dataset.number}")
                hall_symbol = dataset.hall
                if '"' in hall_symbol:
                    hall_symbol = hall_symbol.replace('"', '\\"')
                lines.append(f'  Hall_symbol: "{hall_symbol}"')
                lines.append("")
        return lines

    def _cell_info_yaml_lines(self):
        lines = self._primitive_matrix_yaml_lines(
            self._data.primitive_matrix, "primitive_matrix"
        )
        lines += self._supercell_matrix_yaml_lines(
            self._data.supercell_matrix, "supercell_matrix"
        )
        lines += self._primitive_yaml_lines(self._data.primitive, "primitive_cell")
        lines += self._unitcell_yaml_lines()
        lines += self._supercell_yaml_lines()
        return lines

    def _primitive_matrix_yaml_lines(self, matrix, name):
        lines = []
        if matrix is not None:
            lines.append(f"{name}:")
            for v in matrix:
                lines.append("- [ %18.15f, %18.15f, %18.15f ]" % tuple(v))
            lines.append("")
        return lines

    def _supercell_matrix_yaml_lines(self, matrix, name):
        lines = []
        if matrix is not None:
            lines.append(f"{name}:")
            for v in matrix:
                lines.append("- [ %3d, %3d, %3d ]" % tuple(v))
            lines.append("")
        return lines

    def _primitive_yaml_lines(self, primitive: Primitive, name):
        lines = []
        if primitive is not None:
            lines += self._cell_yaml_lines(self._data.primitive, name, None)
            lines.append("  reciprocal_lattice: # without 2pi")
            rec_lat = np.linalg.inv(primitive.cell)
            for v, a in zip(rec_lat.T, ("a*", "b*", "c*")):
                lines.append(
                    "  - [ %21.15f, %21.15f, %21.15f ] # %s" % (v[0], v[1], v[2], a)
                )
            lines.append("")
        return lines

    def _unitcell_yaml_lines(self):
        lines = []
        if self._data.unitcell is not None:
            if isinstance(self._data.primitive, Primitive) and isinstance(
                self._data.supercell, Supercell
            ):
                s2p_map = self._data.primitive.s2p_map
                u2s_map = self._data.supercell.u2s_map
                u2u_map = self._data.supercell.u2u_map
                s2u_map = self._data.supercell.s2u_map
                u2p_map = [u2u_map[i] for i in (s2u_map[s2p_map])[u2s_map]]
            else:
                u2p_map = None
            lines += self._cell_yaml_lines(self._data.unitcell, "unit_cell", u2p_map)
            lines.append("")
        return lines

    def _supercell_yaml_lines(self):
        lines = []
        if self._data.supercell is not None:
            s2p_map = getattr(self._data.primitive, "s2p_map", None)
            lines += self._cell_yaml_lines(self._data.supercell, "supercell", s2p_map)
            lines.append("")
        return lines

    def _cell_yaml_lines(
        self, cell: PhonopyAtoms, name, map_to_primitive: Optional[list] = None
    ):
        lines = []
        lines.append("%s:" % name)
        count = 0
        for line in cell.get_yaml_lines():
            lines.append("  " + line)
            if map_to_primitive is not None and "mass" in line:
                lines.append("    reduced_to: %d" % (map_to_primitive[count] + 1))
                count += 1
        return lines

    @abstractmethod
    def _nac_yaml_lines(self):
        pass

    def _nac_yaml_lines_given_symbols(self, symbols=None):
        lines = []
        if self._data.nac_params is not None:
            if self._dumper_settings["born_effective_charge"]:
                lines.append("  born_effective_charge:")
                for i, z in enumerate(self._data.nac_params["born"]):
                    text = "  - # %d" % (i + 1)
                    if symbols:
                        text += " (%s)" % symbols[i]
                    lines.append(text)
                    for v in z:
                        lines.append("    - [ %18.15f, %18.15f, %18.15f ]" % tuple(v))
            if self._dumper_settings["dielectric_constant"]:
                lines.append("  dielectric_constant:")
                for v in self._data.nac_params["dielectric"]:
                    lines.append("    - [ %18.15f, %18.15f, %18.15f ]" % tuple(v))
            if lines:
                if "method" in self._data.nac_params:
                    if self._data.nac_params["method"].lower() == "gonze":
                        lines.append('  method: "Gonze"')
                    if self._data.nac_params["method"].lower() == "wang":
                        lines.append('  method: "Wang"')
                if "factor" in self._data.nac_params:
                    factor = self._data.nac_params["factor"]
                    # if factor is not None:
                    lines.append("  unit_conversion_factor: %f" % factor)
                lines.insert(0, "nac:")
                lines.append("")
        return lines

    def _dataset_yaml_lines(self):
        lines = []
        if (
            self._dumper_settings["force_sets"]
            or self._dumper_settings["displacements"]
        ):
            disp_yaml_lines = self._displacements_yaml_lines(
                with_forces=self._dumper_settings["force_sets"]
            )
            lines += disp_yaml_lines
        return lines

    @abstractmethod
    def _displacements_yaml_lines(self, with_forces=False) -> list:
        pass

    def _displacements_yaml_lines_2types(
        self,
        dataset: dict,
        with_forces: bool = False,
        key_prefix: str = "",
        v223_mode: bool = False,
    ) -> list:
        """Choose yaml writer depending on the dataset type.

        See type1 and type2 at Phonopy.dataset.

        Parameters
        ----------
        v223_mode : bool
            When True, old dataset yaml format is generated. Default is False.

        """
        if dataset is not None:
            if "first_atoms" in dataset:
                return self._displacements_yaml_lines_type1(
                    dataset, with_forces=with_forces, key_prefix=key_prefix
                )
            elif "displacements" in dataset:
                if v223_mode:
                    return self._displacements_yaml_lines_type2_v223(
                        dataset, with_forces=with_forces, key_prefix=key_prefix
                    )
                else:
                    return self._displacements_yaml_lines_type2(
                        dataset, with_forces=with_forces, key_prefix=key_prefix
                    )
        return []

    @abstractmethod
    def _displacements_yaml_lines_type1(
        self, dataset: dict, with_forces: bool = False, key_prefix: str = ""
    ) -> list:
        pass

    def _displacements_yaml_lines_type2(
        self, dataset: dict, with_forces: bool = False, key_prefix: str = ""
    ) -> list:
        """Return type2 dataset in yaml.

        This is the format >= v2.24.

        See data structure at Phonopy.dataset.

        """
        lines = [f"{key_prefix}dataset:"]
        if "random_seed" in dataset:
            lines.append("  random_seed: {:d}".format(dataset["random_seed"]))
        for key in ("displacements", "forces"):
            if key not in dataset:
                continue
            if key == "forces" and not with_forces:
                continue
            lines.append(f"  {key}:")
            for i, dset in enumerate(dataset[key]):
                lines.append(f"  - # {i + 1}")
                for _, d in enumerate(dset):
                    lines.append("    - [ %21.16f, %21.16f, %21.16f ]" % tuple(d))

        if "supercell_energies" in dataset:
            lines.append("  supercell_energies:")
            for i, energy in enumerate(dataset["supercell_energies"]):
                lines.append(f"  - {energy:.16f} # {i + 1}")
            lines.append("")

        return lines

    def _displacements_yaml_lines_type2_v223(
        self, dataset: dict, with_forces: bool = False, key_prefix: str = ""
    ) -> list:
        """Return type2 dataset in old stype yaml.

        This is the format < v2.24.

        See data structure at Phonopy.dataset.

        """
        if "random_seed" in dataset:
            lines = [
                "random_seed: %d" % dataset["random_seed"],
                f"{key_prefix}displacements:",
            ]
        else:
            lines = [
                f"{key_prefix}displacements:",
            ]
        for i, dset in enumerate(dataset["displacements"]):
            lines.append(f"- # {i + 1}")
            for j, d in enumerate(dset):
                lines.append(f"  - displacement: # {j + 1}")
                lines.append("      [ %20.16f,%20.16f,%20.16f ]" % tuple(d))
                if with_forces and "forces" in dataset:
                    f = dataset["forces"][i][j]
                    lines.append("    force:")
                    lines.append("      [ %20.16f,%20.16f,%20.16f ]" % tuple(f))
        lines.append("")

        if "supercell_energies" in dataset:
            lines.append(f"{key_prefix}supercell_energies:")
            for i, energy in enumerate(dataset["supercell_energies"]):
                lines.append(f"- {energy:.8f} # {i + 1}")
            lines.append("")

        return lines

    def _force_constants_yaml_lines(self):
        lines = []
        if (
            self._dumper_settings["force_constants"]
            and self._data.force_constants is not None
        ):
            shape = self._data.force_constants.shape[:2]
            lines = [
                "force_constants:",
            ]
            if shape[0] == shape[1]:
                lines.append('  format: "full"')
            else:
                lines.append('  format: "compact"')
            lines.append("  shape: [ %d, %d ]" % shape)
            lines.append("  elements:")
            for i, j in list(np.ndindex(shape)):
                lines.append("  - # (%d, %d)" % (i + 1, j + 1))
                for v in self._data.force_constants[i, j]:
                    lines.append("    - [ %21.15f, %21.15f, %21.15f ]" % tuple(v))
        return lines

    def _init_dumper_settings(self, dumper_settings):
        self._dumper_settings = self._default_dumper_settings.copy()
        if isinstance(dumper_settings, dict):
            self._dumper_settings.update(dumper_settings)


class PhonopyYamlDumper(PhonopyYamlDumperBase):
    """Base class of PhonopyYaml dumper."""

    def _displacements_yaml_lines(self, with_forces=False) -> list:
        return self._displacements_yaml_lines_2types(
            self._data.dataset, with_forces=with_forces
        )

    def _nac_yaml_lines(self):
        if self._data.primitive is None:
            symbols = []
        else:
            symbols = self._data.primitive.symbols
        return self._nac_yaml_lines_given_symbols(symbols)

    def _displacements_yaml_lines_type1(
        self, dataset: dict, with_forces: bool = False, key_prefix: str = ""
    ) -> list:
        """Return type1 dataset in yaml.

        See data structure at Phonopy.dataset.

        """
        return _displacements_yaml_lines_type1(
            dataset, with_forces=with_forces, key_prefix=key_prefix
        )


class PhonopyYaml:
    """PhonopyYaml is a container of phonopy setting.

    This contains the writer (__str__) and reader (read) of phonopy.yaml type
    file.

    Methods
    -------
    __str__
        Return string of phonopy.yaml.
    get_yaml_lines
        Return a list of string lines of phonopy.yaml.
    read
        Read specific properties written in phonopy.yaml.
    set_phonon_info
        Copy specific properties in Phonopy instance to self.

    Attributes
    ----------
    configuration : dict
        Phonopy setting tags or options (e.g., {"DIM": "2 2 2", ...})
    calculator : str
        Force calculator.
    physical_units : dict
        Set of physical units used in this phonon calculation.
    unitcell : PhonopyAtoms
        Unit cell.
    primitive : Primitive
        Primitive cell. The instance of Primitive class is necessary has to
        be created from the instance of Supercell class with
        np.dot(np.linalg.inv(supercell_matrix), primitive_matrix).
    supercell : Supercell
        Supercell. The instance of Supercell class is necessary has to be
        created from unitcell with supercel_matrix.
    dataset
    supercell_matrix
    primitive_matrix
    nac_params
    force_constants
    symmetry
    frequency_unit_conversion_factor
    version
    settings
    command_name
    default_filenames

    """

    default_filenames = ("phonopy_disp.yaml", "phonopy.yaml")
    command_name = "phonopy"

    configuration = phonopy_yaml_property_factory("configuration")
    calculator = phonopy_yaml_property_factory("calculator")
    physical_units = phonopy_yaml_property_factory("physical_units")
    unitcell = phonopy_yaml_property_factory("unitcell")
    primitive = phonopy_yaml_property_factory("primitive")
    supercell = phonopy_yaml_property_factory("supercell")
    dataset = phonopy_yaml_property_factory("dataset")
    supercell_matrix = phonopy_yaml_property_factory("supercell_matrix")
    primitive_matrix = phonopy_yaml_property_factory("primitive_matrix")
    nac_params = phonopy_yaml_property_factory("nac_params")
    force_constants = phonopy_yaml_property_factory("force_constants")
    symmetry = phonopy_yaml_property_factory("symmetry")
    frequency_unit_conversion_factor = phonopy_yaml_property_factory(
        "frequency_unit_conversion_factor"
    )
    version = phonopy_yaml_property_factory("version")

    def __init__(
        self, configuration=None, calculator=None, physical_units=None, settings=None
    ):
        """Init method.

        Parameters
        ----------
        configuration : dict
            Key-value pairs of phonopy calculation settings.
        calculator : str
            Force calculator name.
        physical_units : dict
            Physical units used for the calculation.
        settings : dict
            This controls amount of information in yaml output. See Phonopy.save().

        """
        self._data = PhonopyYamlData(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
        )
        self._dumper_settings = settings

    def __str__(self):
        """Return string text of yaml output."""
        phyml_dumper = PhonopyYamlDumper(
            self._data, dumper_settings=self._dumper_settings
        )
        return "\n".join(phyml_dumper.get_yaml_lines())

    def read(self, filename: Union[str, bytes, os.PathLike, io.IOBase]):
        """Read PhonopyYaml file."""
        self._data = read_phonopy_yaml(
            filename,
            configuration=self._data.configuration,
            calculator=self._data.calculator,
            physical_units=self._data.physical_units,
        )
        return self

    def set_phonon_info(self, phonopy: "Phonopy") -> PhonopyYaml:
        """Collect data from Phonopy instance."""
        self._data.unitcell = phonopy.unitcell
        self._data.primitive = phonopy.primitive
        self._data.supercell = phonopy.supercell
        self._data.version = phonopy.version
        self._data.supercell_matrix = phonopy.supercell_matrix
        self._data.symmetry = phonopy.symmetry
        self._data.primitive_matrix = phonopy.primitive_matrix
        self._data.nac_params = phonopy.nac_params
        self._data.frequency_unit_conversion_factor = phonopy.unit_conversion_factor
        self._data.calculator = phonopy.calculator
        self._data.force_constants = phonopy.force_constants
        self._data.dataset = phonopy.dataset
        return self


def read_phonopy_yaml(
    filename: Union[str, bytes, os.PathLike, io.IOBase],
    configuration=None,
    calculator=None,
    physical_units=None,
) -> PhonopyYamlData:
    """Read phonopy.yaml like file."""
    yaml_data = load_yaml(filename)
    if isinstance(yaml_data, str):
        if isinstance(filename, io.IOBase):
            msg = "Could not load stream properly."
        else:
            msg = f'Could not load "{filename}" properly.'
        raise TypeError(msg)

    return load_phonopy_yaml(
        yaml_data,
        configuration=configuration,
        calculator=calculator,
        physical_units=physical_units,
    )


def load_phonopy_yaml(
    yaml_data, configuration=None, calculator=None, physical_units=None
) -> PhonopyYamlData:
    """Return PhonopyYamlData instance loading yaml data.

    Parameters
    ----------
    yaml_data : dict

    """
    phyml_loader = PhonopyYamlLoader(
        yaml_data,
        configuration=configuration,
        calculator=calculator,
        physical_units=physical_units,
    )
    phyml_loader.parse()
    return phyml_loader.data


def read_cell_yaml(
    filename: Union[str, bytes, os.PathLike, io.IOBase], cell_type: str = "unitcell"
) -> PhonopyAtoms:
    """Read crystal structure from a phonopy.yaml or PhonopyAtoms.__str__ like file.

    phonopy.yaml like file can contain several different cells, e.g., unit cell,
    primitive cell, or supercell. In this case, the default preference order of
    the returned cell is unit cell > primitive cell > supercell. ``cell_type``
    is used to specify to choose one of them.

    When output of PhonopyAtoms.__str__ is given (like below), this file is
    parsed and its cell is returned.

    lattice:
    - [     0.000000000000000,     2.845150738087836,     2.845150738087836 ] # a
    - [     2.845150738087836,     0.000000000000000,     2.845150738087836 ] # b
    - [     2.845150738087836,     2.845150738087836,     0.000000000000000 ] # c
    points:
    - symbol: Na # 1
      coordinates: [  0.000000000000000,  0.000000000000000,  0.000000000000000 ]
      mass: 22.989769
    - symbol: Cl # 2
      coordinates: [  0.500000000000000,  0.500000000000000,  0.500000000000000 ]
      mass: 35.453000

    Parameters
    ----------
    filename : str, bytes, os.PathLike, io.IOBase
        File name or file stream.
    cell_type : str
        "unitcell", "primitive", or "supercell". Default is "unitcell".

    """
    ph_yaml = PhonopyYaml()
    ph_yaml.read(filename)
    if ph_yaml.unitcell and cell_type == "unitcell":
        return ph_yaml.unitcell
    elif ph_yaml.primitive and cell_type == "primitive":
        return ph_yaml.primitive
    elif ph_yaml.supercell and cell_type == "supercell":
        return ph_yaml.supercell
    else:
        if isinstance(filename, io.IOBase):
            raise RuntimeError("Crystal structure data was not found in stream.")
        else:
            raise RuntimeError(f'Crystal structure data was not found in "{filename}".')


def load_yaml(fp: Union[str, bytes, os.PathLike, io.IOBase]):
    """Load yaml file.

    Parameters
    ----------
    fp : str, bytes, os.PathLike or io.IOBase
        Filename, file path, or file stream.

    lzma and gzip comppressed non-stream files can be loaded.

    """
    if isinstance(fp, io.IOBase):
        yaml_data = yaml.load(fp, Loader=Loader)
    else:
        myio = get_io_module_to_decompress(fp)
        with myio.open(fp) as f:
            yaml_data = yaml.load(f, Loader=Loader)

    return yaml_data


def _displacements_yaml_lines_type1(
    dataset: dict, with_forces: bool = False, key_prefix: str = ""
) -> list:
    """Return type1 dataset in yaml.

    See data structure at Phonopy.dataset.

    """
    lines = [
        f"{key_prefix}displacements:",
    ]
    for d in dataset["first_atoms"]:
        lines.append("- atom: %4d" % (d["number"] + 1))
        lines.append("  displacement:")
        lines.append("    [ %20.16f,%20.16f,%20.16f ]" % tuple(d["displacement"]))
        if with_forces and "forces" in d:
            lines.append("  forces:")
            for f in d["forces"]:
                lines.append("  - [ %20.16f,%20.16f,%20.16f ]" % tuple(f))
        if "supercell_energy" in d:
            lines.append(
                "  supercell_energy: {energy:.8f}".format(energy=d["supercell_energy"])
            )
    lines.append("")
    return lines



# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

import numpy as np

from phonopy import PhonopyQHA
from phonopy.file_IO import read_efe, read_thermal_properties_yaml, read_v_e
from phonopy.physical_units import get_physical_units


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy-QHA command-line-tool")
    parser.set_defaults(
        pressure=None,
        is_graph_plot=False,
        is_graph_save=False,
        is_bulk_modulus_only=False,
        efe_file=None,
        eos="vinet",
        thin_number=10,
        tmax=1000.0,
    )
    parser.add_argument(
        "-b",
        dest="is_bulk_modulus_only",
        action="store_true",
        help="Just show Bulk modulus from v-e data",
    )
    parser.add_argument(
        "--eos",
        dest="eos",
        help="Choise of EOS among vinet, birch_murnaghan, and murnaghan",
    )
    parser.add_argument(
        "--exclude_imaginary",
        dest="exclude_imaginary",
        action="store_true",
        help="Exclude volumes that show imaginary modes",
    )
    parser.add_argument(
        "-p", "--plot", dest="is_graph_plot", action="store_true", help="Plot data"
    )
    parser.add_argument(
        "--pressure", dest="pressure", type=float, help="Pressure in GPa"
    )
    parser.add_argument(
        "--efe",
        "--electronic-free-energy",
        dest="efe_file",
        nargs=1,
        help="Read electronic free energies at temperatures and volumes",
    )
    parser.add_argument(
        "-s",
        "--save",
        dest="is_graph_save",
        action="store_true",
        help="Save plot data in pdf",
    )
    parser.add_argument(
        "--sparse",
        dest="thin_number",
        type=int,
        help=(
            "Thin out the F-V plots of temperature. The value is "
            "used as deviser of number of temperature points."
        ),
    )
    parser.add_argument(
        "--tmax", dest="tmax", type=float, help="Maximum calculated temperature"
    )
    parser.add_argument(
        "filenames",
        nargs="*",
        help="Filenames of e-v.dat and thermal_properties.yaml's",
    )
    args = parser.parse_args()
    return args


def run():
    """Run phonopy-qha."""
    args = get_options()

    if args.is_graph_save:
        import matplotlib

        matplotlib.use("Agg")
        matplotlib.rc("pdf", fonttype=42)

    # Choose EOS
    if args.eos == "birch_murnaghan":
        print("# Third-order Birch-Murnaghan EOS")
    elif args.eos == "murnaghan":
        print("# Murnaghan EOS")
    else:
        print("# Vinet EOS")

    ########################
    # Read data from files #
    ########################
    volumes, electronic_energies = read_v_e(args.filenames[0])
    if args.efe_file:
        _temperatures, electronic_energies = read_efe(args.efe_file[0])
    # electronic_energies += volumes * args.pressure / EVAngstromToGPa

    # Show bulk modulus of v-e data
    if args.is_bulk_modulus_only:
        # if args.efe_file:
        #     print("--efe optin can't be used with -b option")
        #     sys.exit(1)
        bulk_modulus = PhonopyQHA(
            volumes,
            electronic_energies=electronic_energies,
            pressure=args.pressure,
            eos=args.eos,
        )
        parameters = bulk_modulus.get_bulk_modulus_parameters()
        if args.efe_file:
            print(f"Volume: {parameters[3]}")
            print(f"Energy: {parameters[0]}")
            print(
                f"Bulk modulus: {parameters[1] * get_physical_units().EVAngstromToGPa}"
            )
            print(
                f"Parameters: {parameters[0]} {parameters[1]} "
                f"{parameters[2]} {parameters[3]}"
            )
        else:
            print(f"Volume: {parameters[3]:.7f}")
            print(f"Energy: {parameters[0]:.7f}")
            print(
                "Bulk modulus: "
                f"{parameters[1] * get_physical_units().EVAngstromToGPa:.7f}"
            )
            print(
                f"Parameters: {parameters[0]:.7f} {parameters[1]:.7f} "
                f"{parameters[2]:.7f} {parameters[3]:.7f}"
            )

        if args.is_graph_plot:
            bulk_modulus.plot_bulk_modulus(thin_number=args.thin_number).show()

        # These lines commented out are to print E-V points on the curve fitted.
        # def eos(v):
        #     params = bulk_modulus._bulk_modulus.get_parameters()
        #     _eos = bulk_modulus._bulk_modulus.get_eos()
        #     return _eos(v, *params)

        # print("E-V values")
        # for v in volumes:
        #     print(v, eos(v))

        sys.exit(0)

    # Check number of files in e-v.dat case
    if len(volumes) != len(args.filenames[1:]):
        print(
            "The number of thermal_properites.yaml files (%d) "
            "is inconsisten with" % len(args.filenames[1:])
        )
        print("the number of e-v data (%d)." % len(volumes))
        sys.exit(1)

    if args.efe_file:
        if len(volumes) != electronic_energies.shape[1]:
            print(
                "%s and %s are inconsistent for the volume points."
                % (args.filenames[0], args.efe_file[0])
            )
            sys.exit(1)

    (
        temperatures,
        cv,
        entropy,
        fe_phonon,
        num_modes,
        num_integrated_modes,
    ) = read_thermal_properties_yaml(args.filenames[1:])

    if args.efe_file:
        if (
            len(temperatures) >= len(_temperatures)
            and (
                np.abs(temperatures[: len(_temperatures)] - _temperatures) > 1e-5
            ).any()
        ) or (
            len(temperatures) < len(_temperatures)
            and (np.abs(temperatures - _temperatures[: len(temperatures)]) > 1e-5).any()
        ):
            print(
                f"Inconsistency is found in temperatures in {args.efe_file[0]} and "
                "thermal_properties.yaml files."
            )
            print("Temperatures in a thermal_properties.yaml are")
            print(f"{temperatures}.")
            print(f"Temperatures in {args.efe_file[0]} are")
            print(f"{_temperatures.tolist()}.")
            sys.exit(1)

    ########################################################
    # Treatment of thermal properties with imaginary modes #
    ########################################################
    if args.exclude_imaginary and num_modes:
        indices = []
        num_imag_modes = np.array(num_modes) - np.array(num_integrated_modes)
        for i, nim in enumerate(num_imag_modes):
            if nim < 4:
                indices.append(i)
    else:
        indices = range(len(volumes))

    if args.efe_file:
        electronic_energies = electronic_energies[:, indices]
    else:
        electronic_energies = electronic_energies[indices]

    ##########################
    # Analyzing and plotting #
    ##########################
    phonopy_qha = PhonopyQHA(
        volumes=volumes[indices],
        electronic_energies=electronic_energies,
        eos=args.eos,
        temperatures=temperatures,
        pressure=args.pressure,
        free_energy=fe_phonon[:, indices],
        cv=cv[:, indices],
        entropy=entropy[:, indices],
        t_max=args.tmax,
        verbose=True,
    )

    if num_modes:
        num_imag_modes = np.array(num_modes) - np.array(num_integrated_modes)
        for filename, nim in zip(
            args.filenames[1 : (len(volumes) + 1)], num_imag_modes
        ):
            if nim > 3:
                if args.exclude_imaginary:
                    print("# %s has been excluded." % filename)
                else:
                    print("# Warning: %s has imaginary modes." % filename)

    if args.is_graph_plot and not args.is_graph_save:
        # Plot on display
        # - Volume vs Helmholtz free energy
        # - Volume vs Temperature
        # - Thermal expansion coefficient
        phonopy_qha.plot_qha(thin_number=args.thin_number).show()

    if args.is_graph_save:
        # Volume vs Helmholts free energy
        phonopy_qha.plot_pdf_helmholtz_volume(thin_number=args.thin_number)

        # Volume vs Temperature
        phonopy_qha.plot_pdf_volume_temperature()

        # Thermal expansion coefficient
        phonopy_qha.plot_pdf_thermal_expansion()

        # G vs Temperature
        phonopy_qha.plot_pdf_gibbs_temperature()

        # Bulk modulus vs Temperature
        phonopy_qha.plot_pdf_bulk_modulus_temperature()

        # C_P vs Temperature
        phonopy_qha.plot_pdf_heat_capacity_P_numerical()

        # C_P vs Temperature (poly fit)
        phonopy_qha.plot_pdf_heat_capacity_P_polyfit()

        # Gruneisen parameter vs Temperature
        phonopy_qha.plot_pdf_gruneisen_temperature()

    phonopy_qha.write_helmholtz_volume()
    phonopy_qha.write_helmholtz_volume_fitted(thin_number=args.thin_number)
    phonopy_qha.write_volume_temperature()
    phonopy_qha.write_thermal_expansion()
    phonopy_qha.write_gibbs_temperature()
    phonopy_qha.write_bulk_modulus_temperature()
    phonopy_qha.write_heat_capacity_P_numerical()
    phonopy_qha.write_heat_capacity_P_polyfit()
    phonopy_qha.write_gruneisen_temperature()
